/************************************************************************************

Filename    :   DebugLines.cpp
Content     :   Class that manages and renders debug lines.
Created     :   April 22, 2014
Authors     :   Jonathan E. Wright

Copyright   :   Copyright (c) Facebook Technologies, LLC and its affiliates. All rights reserved.

*************************************************************************************/

#include "DebugLines.h"

#include <stdlib.h>

#include "OVR_GlUtils.h"
#include "Kernel/OVR_LogUtils.h"

#include "GlGeometry.h"
#include "GlProgram.h"

namespace OVR {

static const char * DebugLineVertexSrc =
	"attribute vec4 Position;\n"
	"attribute vec4 VertexColor;\n"
	"varying lowp vec4 outColor;\n"
	"void main()\n"
	"{\n"
	"   gl_Position = TransformVertex( Position );\n"
	"   outColor = VertexColor;\n"
	"}\n";

static const char * DebugLineFragmentSrc =
	"varying lowp vec4 outColor;\n"
	"void main()\n"
	"{\n"
	"	gl_FragColor = outColor;\n"
	"}\n";

//==============================================================
// OvrDebugLinesLocal
//
class OvrDebugLinesLocal : public OvrDebugLines
{
public:
	struct DebugLines_t
	{
		DebugLines_t()
		: DrawSurf( & Surf )
		{
		}

		ovrSurfaceDef 		Surf;
		ovrDrawSurface 		DrawSurf;
		VertexAttribs 		Attr;
		Array<long long> 	EndFrame;
	};

	typedef unsigned short LineIndex_t;

	static const int MAX_DEBUG_LINES = 2048;

						OvrDebugLinesLocal();
	virtual				~OvrDebugLinesLocal();

	virtual void		Init();
	virtual void		Shutdown();

	virtual void		BeginFrame( const long long frameNum );
	virtual void		AppendSurfaceList( Array<ovrDrawSurface> & surfaceList );

	virtual void		AddLine(	const Vector3f & start, const Vector3f & end,
								const Vector4f & startColor, const Vector4f & endColor,
								const long long endFrame, const bool depthTest );
	virtual void		AddPoint(	const Vector3f & pos, const float size,
								const Vector4f & color, const long long endFrame,
								const bool depthTest );
	// Add a debug point without a specified color. The axis lines will use default
	// colors: X = red, Y = green, Z = blue (same as Maya).
	virtual void		AddPoint(	const Vector3f & pos, const float size,
								const long long endFrame, const bool depthTest );

	virtual void		AddBounds( Posef const & pose, Bounds3f const & bounds, Vector4f const & color );


	virtual void		    AddAxes( const Vector3f & origin, const Matrix4f & axes, const float size, 
						    		const Vector4f & color, const long long endFrame, 
						    		const bool depthTest );


private:
	DebugLines_t		DepthTested;
	DebugLines_t		NonDepthTested;

	bool				Initialized;
	GlProgram			LineProgram;

	void		RemoveExpired( const long long frameNum, DebugLines_t & lines );
};

//==============================
// OvrDebugLinesLocal::OvrDebugLinesLocal
OvrDebugLinesLocal::OvrDebugLinesLocal() :
	Initialized( false )
{
}

//==============================
// OvrDebugLinesLocal::OvrDebugLinesLocal
OvrDebugLinesLocal::~OvrDebugLinesLocal()
{
	Shutdown();
}

//==============================
// OvrDebugLinesLocal::Init
void OvrDebugLinesLocal::Init()
{
	if ( Initialized )
	{
// JDC: multi-activity test		ASSERT_WITH_TAG( !Initialized, "DebugLines" );
		return;
	}

	// this is only freed by the OS when the program exits
	if ( LineProgram.VertexShader == 0 || LineProgram.FragmentShader == 0 )
	{
		LineProgram = GlProgram::Build( DebugLineVertexSrc, DebugLineFragmentSrc, NULL, 0 );
	}

	// the indices will never change once we've set them up, we just won't necessarily
	// use all of the index buffer to render.
	const int MAX_INDICES = MAX_DEBUG_LINES * 2;
	Array<LineIndex_t> indices;
	indices.Reserve( MAX_INDICES );

	for ( LineIndex_t i = 0; i < MAX_INDICES; ++i )
	{
		indices.PushBack(i);
	}

	for ( int i = 0; i < 2; i++ )
	{
		DebugLines_t & dl = i == 0 ? NonDepthTested : DepthTested;
		dl.Surf.geo.Create( dl.Attr, indices );
		dl.Surf.geo.primitiveType = GL_LINES;
		ovrGraphicsCommand & gc = dl.Surf.graphicsCommand;
		gc.GpuState.blendDst = GL_ONE_MINUS_SRC_ALPHA;
		gc.GpuState.depthEnable = gc.GpuState.depthMaskEnable = i == 1;
		gc.GpuState.lineWidth = 2.0f;
		gc.Program = LineProgram;
	}

	Initialized = true;
}

//==============================
// OvrDebugLinesLocal::Shutdown
void OvrDebugLinesLocal::Shutdown()
{
	if ( !Initialized )
	{
		OVR_ASSERT_WITH_TAG( !Initialized, "DebugLines" );
		return;
	}
	DepthTested.Surf.geo.Free();
	NonDepthTested.Surf.geo.Free();
	GlProgram::Free( LineProgram );
	Initialized = false;
}

//==============================
// OvrDebugLinesLocal::AppendSurfaceList
void OvrDebugLinesLocal::AppendSurfaceList( Array< ovrDrawSurface > & surfaceList )
{
	for ( int j = 0; j < 2; j++ )
	{
		DebugLines_t & dl = j == 0 ? NonDepthTested : DepthTested;
		int verts = dl.Attr.position.GetSizeI();
		if ( verts == 0 )
		{
			continue;
		}
		dl.Surf.geo.Update( dl.Attr );
		dl.Surf.geo.indexCount = verts;
		surfaceList.PushBack( dl.DrawSurf );
	}
}


//==============================
// OvrDebugLinesLocal::AddLine
void OvrDebugLinesLocal::AddLine( const Vector3f & start, const Vector3f & end,
		const Vector4f & startColor, const Vector4f & endColor,
		const long long endFrame, const bool depthTest )
{
	//OVR_LOG( "OvrDebugLinesLocal::AddDebugLine" );
	DebugLines_t & dl = depthTest ? DepthTested : NonDepthTested;
	dl.Attr.position.PushBack( start );
	dl.Attr.position.PushBack( end );
	dl.Attr.color.PushBack( startColor );
	dl.Attr.color.PushBack( endColor );
	dl.EndFrame.PushBack( endFrame );
    //OVR_ASSERT( DepthTested.EndFrame.GetSizeI() < MAX_DEBUG_LINES );
    //OVR_ASSERT( NonDepthTested.EndFrame.GetSizeI() < MAX_DEBUG_LINES );
}

//==============================
// OvrDebugLinesLocal::AddPoint
void OvrDebugLinesLocal::AddPoint(	const Vector3f & pos, const float size, const Vector4f & color,
		const long long endFrame, const bool depthTest )
{
	float const hs = size * 0.5f;
	Vector3f const fwd( 0.0f, 0.0f, hs );
	Vector3f const right( hs, 0.0f, 0.0f );
	Vector3f const up( 0.0f, hs, 0.0f );

	AddLine( pos - fwd, pos + fwd, color, color, endFrame, depthTest );
	AddLine( pos - right , pos + right, color, color, endFrame, depthTest );
	AddLine( pos - up, pos + up, color, color, endFrame, depthTest );
}

//==============================
// OvrDebugLinesLocal::AddPoint
void OvrDebugLinesLocal::AddPoint(	const Vector3f & pos, const float size,
		const long long endFrame, const bool depthTest )
{
	float const hs = size * 0.5f;
	Vector3f const fwd( 0.0f, 0.0f, hs );
	Vector3f const right( hs, 0.0f, 0.0f );
	Vector3f const up( 0.0f, hs, 0.0f );

	AddLine( pos - fwd, pos + fwd, Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ), Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ), endFrame, depthTest );
	AddLine( pos - right, pos + right, Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ), Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ), endFrame, depthTest );
	AddLine( pos - up, pos + up, Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ), Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ), endFrame, depthTest );
}

//==============================
// OvrDebugLinesLocal::AddAxes
void OvrDebugLinesLocal::AddAxes( const Vector3f & origin, const Matrix4f & axes, const float size, 
						    		const Vector4f & color, const long long endFrame, 
						    		const bool depthTest )
{
	const float half_size = size * 0.5f;
	Vector3f const fwd		= Vector3f( axes.M[2][0], axes.M[2][1], axes.M[2][2] ) * half_size;
	Vector3f const right	= Vector3f( axes.M[0][0], axes.M[0][1], axes.M[0][2] ) * half_size;
	Vector3f const up		= Vector3f( axes.M[1][0], axes.M[1][1], axes.M[1][2] ) * half_size;

	AddLine( origin - fwd,	 origin + fwd,   Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ), Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ), endFrame, depthTest );
	AddLine( origin - right, origin + right, Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ), Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ), endFrame, depthTest );
	AddLine( origin - up,    origin + up,    Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ), Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ), endFrame, depthTest );
}

//==============================
// OvrDebugLinesLocal::AddBounds
void OvrDebugLinesLocal::AddBounds( Posef const & pose, Bounds3f const & bounds, Vector4f const & color )
{
	Vector3f const & mins = bounds.GetMins();
	Vector3f const & maxs = bounds.GetMaxs();
	Vector3f corners[8];
	corners[0] = mins;
	corners[7] = maxs;
	corners[1] = Vector3f( mins.x, maxs.y, mins.z );
	corners[2] = Vector3f( mins.x, maxs.y, maxs.z );
	corners[3] = Vector3f( mins.x, mins.y, maxs.z );
	corners[4] = Vector3f( maxs.x, mins.y, mins.z );
	corners[5] = Vector3f( maxs.x, maxs.y, mins.z );
	corners[6] = Vector3f( maxs.x, mins.y, maxs.z );

	// transform points
	for ( int i = 0; i < 8; ++i )
	{
		corners[i] = pose.Rotation.Rotate( corners[i] );
		corners[i] += pose.Translation;
	}

	AddLine( corners[0], corners[1], color, color, 1, true );
	AddLine( corners[1], corners[2], color, color, 1, true );
	AddLine( corners[2], corners[3], color, color, 1, true );
	AddLine( corners[3], corners[0], color, color, 1, true );
	AddLine( corners[7], corners[6], color, color, 1, true );
	AddLine( corners[6], corners[4], color, color, 1, true );
	AddLine( corners[4], corners[5], color, color, 1, true );
	AddLine( corners[5], corners[7], color, color, 1, true );
	AddLine( corners[0], corners[4], color, color, 1, true );
	AddLine( corners[1], corners[5], color, color, 1, true );
	AddLine( corners[2], corners[7], color, color, 1, true );
	AddLine( corners[3], corners[6], color, color, 1, true );
}

//==============================
// OvrDebugLinesLocal::BeginFrame
void OvrDebugLinesLocal::BeginFrame( const long long frameNum )
{
	// LOG( "OvrDebugLinesLocal::RemoveExpired: frame %lli, removing %i lines", frameNum, DepthTestedLines.GetSizeI() + NonDepthTestedLines.GetSizeI() );
	DepthTested.Surf.geo.indexCount = 0;
	NonDepthTested.Surf.geo.indexCount = 0;
	RemoveExpired( frameNum, DepthTested );
	RemoveExpired( frameNum, NonDepthTested );
}

//==============================
// OvrDebugLinesLocal::RemoveExpired
void OvrDebugLinesLocal::RemoveExpired( const long long frameNum, DebugLines_t & lines )
{
	for ( int i = lines.EndFrame.GetSizeI() - 1; i >= 0; --i )
	{
		if ( frameNum >= lines.EndFrame[i] )
		{
			lines.Attr.position.RemoveMultipleAt( i * 2, 2 );
			lines.Attr.color.RemoveMultipleAt( i * 2, 2 );
			lines.EndFrame.RemoveAt( i );
		}
	}
}

//==============================
// OvrDebugLines::Create
OvrDebugLines * OvrDebugLines::Create()
{
    return new OvrDebugLinesLocal;
}

//==============================
// OvrDebugLines::Free
void OvrDebugLines::Free( OvrDebugLines * & debugLines )
{
    if ( debugLines != NULL )
    {
        delete debugLines;
        debugLines = NULL;
    }
}

} // namespace OVR
